////////////////////////////////////////////////////////////////////
import {vec3, vec4} from 'gl-matrix';
// wfundament /////////////////////////////////////////////////////
import {
  WKeyboardEventHandler,
  WMouseEventHandler,
  WTouchEventHandler,
  wVec4Colors,
}                   from 'wfundament';
// wgtek ///////////////////////////////////////////////////////////
import hasPassive   from 'has-passive-events';

////////////////////////////////////////////////////////////////////
class GameElement {
  constructor(pos, color, speed) {
    this.pos_ = pos;
    this.color_ = color;
    this.speed_ = Math.max(2.0, speed);
    this.alive_ = false;
  }

  reset() {
    vec3.copy(
      this.pos_,
      vec3.fromValues(Math.random() * 60.0 - 30.0, 0.2, -60.0)
    );
    vec4.copy(
      this.color_,
      vec4.fromValues(Math.random(), Math.random(), Math.random(), 1.0)
    );
    this.speed_ = Math.max(2.0, Math.random() * 4.0);
    this.alive_ = false;
  }
}

// app /////////////////////////////////////////////////////////////
export class WgApp {
  constructor(canvas) {
    // Create web gl context
    const gl = canvas.getContext("webgl2");
    if (gl === undefined) {
      window.alert("No webgl 2.0 support");
      return;
    }
    this.glContext_ = new WgContext(gl, canvas);
    this.gl = gl;
    // Create the orbit camera
    this.orbitCamera_ = new WOrbitCamera("OrbitCamera");
    this.orbitCamera_.setViewport(
      this.glContext_.getCanvasWidth(),
      this.glContext_.getCanvasHeight()
    );
    this.orbitCamera_.setPerspective(70.0);
    this.orbitCamera_.setCameraPosition([0.0, 35.0, 60.0]);
    // Primitive render
    this.primRender_ = new WgPrimRender("PrimitiveRender", this.glContext_);
    // Event handlers
    this.score_ = 0;
    this.numberOfElementsSpawned_ = 0;
    this.WKeyboardEventHandler_ = new WKeyboardEventHandler();
    this.mouseEventHander_ = new WMouseEventHandler();
    this.touchEventHandler_ = new WTouchEventHandler();
    this.setupEventHandlers(canvas);
    // Game objects
    this.padPosition_ = vec3.fromValues(0.0, 1.0, 40.0);
    this.padDimension_ = vec3.fromValues(7.0, 2.0, 1.5);
    this.padSpeed_ = 8.0;
    // Game elements
    this.elements_ = [];
    for (let i = 0; i < 16; i++) {
      var gameElement = new GameElement(
        vec3.fromValues(
          Math.random() * 60.0 - 30.0,
          0.2,
          -50.0 - Math.random() * 30.0
        ),
        vec4.fromValues(Math.random(), Math.random(), Math.random(), 1.0),
        Math.random() * 3.0
      );
      this.elements_.push(gameElement);
    }
    // Event loop
    this.lastTimeStamp_ = 0;
    requestAnimationFrame((timestamp) => {
      this.loop(timestamp);
    });
  }

  getWKeyboardEventHandler() {
    return this.WKeyboardEventHandler_;
  }

  getMouseEventHandler() {
    return this.mouseEventHander_;
  }

  getTouchEventHandler() {
    return this.touchEventHandler_;
  }

  setupEventHandlers(canvas) {
    // Keyboard
    window.addEventListener("keydown", (event) => {
      this.getWKeyboardEventHandler().keyDown(event.key);
    });
    window.addEventListener("keyup", (event) => {
      this.getWKeyboardEventHandler().keyUp(event.key);
    });
    // Mouse
    canvas.addEventListener("mousedown", (event) => {
      this.getMouseEventHandler().onMouseDown(event);
    });
    canvas.addEventListener("mouseup", (event) => {
      this.getMouseEventHandler().onMouseUp(event);
    });
    canvas.addEventListener("mousemove", (event) => {
      this.getMouseEventHandler().onMouseMove(event);
    });
    canvas.addEventListener("mouseleave", (event) => {
      this.getMouseEventHandler().onMouseLeave(event);
    });
    canvas.addEventListener("click", (event) => {
      this.getMouseEventHandler().onMouseClick(event);
    });
    canvas.addEventListener("dblclick", (event) => {
      this.getMouseEventHandler().onMouseDoubleClick(event);
    });
    canvas.addEventListener("wheel", (event) => {
      event.preventDefault();
      this.getMouseEventHandler().onMouseWheel(event);
    });
    // Touch
    canvas.addEventListener(
      "touchstart",
      (event) => {
        this.getTouchEventHandler().onTouchStart(event);
      },
      hasPassive ? { passive: false } : false
    );
    canvas.addEventListener(
      "touchend",
      (event) => {
        this.getTouchEventHandler().onTouchEnd(event);
      },
      hasPassive ? { passive: false } : false
    );
    canvas.addEventListener(
      "touchcancel",
      (event) => {
        this.getTouchEventHandler().onTouchCancel(event);
      },
      hasPassive ? { passive: false } : false
    );
    canvas.addEventListener(
      "touchmove",
      (event) => {
        this.getTouchEventHandler().onTouchMove(event);
      },
      hasPassive ? { passive: false } : false
    );
    canvas.addEventListener(
      "contextmenu",
      (e) => {
        e.preventDefault();
      },
      false
    );
  }

  loop(timestamp) {
    let deltaTime = timestamp - this.lastTimeStamp_;
    this.update(deltaTime / 100);
    this.render(timestamp);
    this.lastTimeStamp_ = timestamp;
    requestAnimationFrame((ts) => {
      this.loop(ts);
    });
  }

  update(deltaTime) {
    this.orbitCamera_.update(
      deltaTime,
      this.getMouseEventHandler(),
      this.getWKeyboardEventHandler(),
      this.getTouchEventHandler()
    );
    if (this.getWKeyboardEventHandler().isDown("ArrowLeft")) {
      this.padPosition_[0] -= this.speed_ * deltaTime;
    }
    if (this.getWKeyboardEventHandler().isDown("ArrowRight")) {
      this.padPosition_[0] += this.speed_ * deltaTime;
    }
    if (this.getWKeyboardEventHandler().isDown("ArrowUp")) {
      this.padPosition_[2] -= this.speed_ * deltaTime;
      this.padPosition_[2] = Math.max(this.padPosition_[2], 25.0);
    }
    if (this.getWKeyboardEventHandler().isDown("ArrowDown")) {
      this.padPosition_[2] += this.speed_ * deltaTime;
      this.padPosition_[2] = Math.min(this.padPosition_[2], 40.0);
    }
    // update falling elements
    for (let i = 0; i < this.elements_.length; i++) {
      let elem = this.elements_[i];
      if (!elem.alive_) {
        if (Math.random() * deltaTime < 0.0001) {
          elem.alive_ = true;
          this.numberOfElementsSpawned_++;
        }
      } else {
        elem.pos_[2] += elem.speed_ * deltaTime;
        if (elem.pos_[2] > 80.0) {
          elem.reset();
        }
        // collision with the pad
        if (
          Math.abs(this.padPosition_[0] - elem.pos_[0]) <
            this.padDimension_[0] * 0.5 + 1.0 &&
          Math.abs(this.padPosition_[2] - elem.pos_[2]) <
            this.padDimension_[2] * 0.5 + 1.0
        ) {
          this.score_++;
          elem.reset();
        }
      }
    }
    // reset game
    if (this.numberOfElementsSpawned_ > 64) {
      for (let i = 0; i < this.elements_.length; i++) {
        this.elements_[i].reset();
      }
      this.score_ = 0;
      this.numberOfElementsSpawned_ = 0;
    }
    this.getMouseEventHandler().flush();
    this.getTouchEventHandler().flush();
  }

  render() {
    const gl = this.gl;
    this.primitiveRenderer_.clearPrimitive();
    this.primitiveRenderer_.addGridXZ(10.0, 10.0);
    let vp = this.orbitCamera_.getViewProjection();
    this.primitiveRenderer_.addWireBox(
      this.padPosition_,
      this.padDimension_,
      wVec4Colors.Orange
    );
    for (let i = 0; i < this.elements_.length; i++) {
      if (!this.elements_[i].alive_) {
        continue;
      }
      this.primitiveRenderer_.addWireBox(
        this.elements_[i].pos_,
        vec3.fromValues(1.0, 1.0, 1.0),
        this.elements_[i].color_
      );
    }
    const offset = -this.score_ * 0.5;
    for (let i = 0; i < this.score_; i++) {
      let v0 = vec4.fromValues(offset + i, 15.0, -30.0, 1.0);
      let v1 = vec4.fromValues(offset + i, 20.0, -30.0, 1.0);
      this.primitiveRenderer_.addLine(
        v0,
        wVec4Colors.LightGrey,
        v1,
        wVec4Colors.White
      );
    }
    gl.clearColor(0.3, 0.3, 0.3, 1.0);
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
    gl.enable(gl.DEPTH_TEST);
    this.primitiveRenderer_.render(vp);
  }

  resize(width, height) {
    this.glContext_.canvas_.width = width;
    this.glContext_.canvas_.height = height;
    this.orbitCamera_.setViewport(
      this.glContext_.getCanvasWidth(),
      this.glContext_.getCanvasHeight()
    );
    this.orbitCamera_.setPerspective(70.0);
    this.glContext_.gl_.viewport(0, 0, width, height);
  }
}
